# pylint: disable=missing-docstring,no-self-use,line-too-long

import unittest
from unittest.mock import patch,  MagicMock
from typing import NamedTuple, Optional, Dict
from json import dumps
import soil


class MockHttpResponse(NamedTuple):
    ''' Soil configuration class '''
    status_code: int
    text: str


# pylint: disable=unused-argument
def mock_http_post(url: str,
                   headers: Optional[Dict[str, str]] = None,
                   json: Optional[Dict[str, str]] = None
                   ) -> MockHttpResponse:
    assert json is not None
    response = 'ok'
    return MockHttpResponse(status_code=201, text=dumps(response))


class TestAlerts(unittest.TestCase):

    @patch('soil.api.requests.post', side_effect=mock_http_post)
    def test_create_event(self, mock_post: MagicMock) -> None:
        soil.alerts.event('test_key', 'test_value')
        mock_post.assert_called_once_with(
            'http://test_host.test/v2/alerts/events/',
            headers={'Authorization': 'Bearer test_token'},
            json={'key': 'test_key', 'value': 'test_value'})

    @patch('soil.api.requests.post', side_effect=mock_http_post)
    def test_create_alert(self, mock_post: MagicMock) -> None:
        alert_config = {
            "role": "test_alerts",
            "event_id": "test_event1",
            "condition": {
                "type": "any"
            }
        }
        soil.alerts.alert(alert_config)
        mock_post.assert_called_once_with(
            'http://test_host.test/v2/alerts/alerts/',
            headers={'Authorization': 'Bearer test_token'},
            json=alert_config)



# pylint: disable=missing-module-docstring, missing-class-docstring, missing-function-docstring, no-self-use
import unittest
import pandas as pd
import numpy as np
from unittest.mock import patch
from sklearn.ensemble import RandomForestClassifier
from soil.data_structures.random_forest_model import RandomForestModel
from iris_dataset.modules import retrain_predictor
from .patch_decorators import patch_modulify


metadata = {
    "_id": "_id",
    "columns": {
        "X_columns": {
            "sepal_length": {
                "name": "sepal_length",
                        "type": "float"
            },
            "sepal_width": {
                "name": "sepal_width",
                        "type": "float"
            },
            "petal_length": {
                "name": "petal_length",
                        "type": "float"
            },
            "petal_width": {
                "name": "petal_width",
                        "type": "float"
            }
        },
        "y_columns": {
            "species": {
                "name": "species",
                        "type": "str"
            }
        }
    }
}


class InputDS:
    def __init__(self, data, metadata=None):
        self.data = data
        self.metadata = metadata if metadata is not None else {}


class TestModulifiedRetrainPredictor(unittest.TestCase):
    def test_output_types(self):
        # This is a bit hacky but in tests it's fine.
        output_types_fn = retrain_predictor.retrain_predictor.__closure__[-1].cell_contents
        assert output_types_fn() == [RandomForestModel]


@patch_modulify(retrain_predictor)
class TestRetrainPredictor(unittest.TestCase):
    @patch('iris_dataset.modules.retrain_predictor.RandomForestModel')
    def test_retrain_predictor(self, mocked_rfmodel):
        np.random.seed(1111)
        mocked_rfmodel.return_value = 'ok'
        fixture = pd.read_csv('./data/iris.csv')
        dataset = InputDS(fixture, metadata)
        [result] = retrain_predictor.retrain_predictor(dataset)
        assert result == 'ok'
        assert len(mocked_rfmodel.call_args_list) == 1
        args, _kwargs = mocked_rfmodel.call_args_list[0]
        assert len(args) == 2
        assert type(args[0]) == RandomForestClassifier
        assert args[0].oob_score
        assert args[1] == {
            'oob_score': 0.9466666666666667,
            'error': 0.053333333333333344,
            'x_labels': metadata['columns']['X_columns'],
            'y_labels': {'species': {'name': 'species', 'type': 'str'}}

        }



# pylint: disable=missing-docstring,no-self-use,line-too-long

from typing import Any, Dict, Optional
import unittest
from unittest.mock import MagicMock, patch
from json import loads, dumps
from soil.dictionary import dictionary


TEST_DICT = {
    "_id": '609bdfe9f5e2c907fa353bca99',
    "content": {
      "a": 'b'
    },
    "created_at": 1620828137994,
    "language": 'es',
    "name": 'testdicaaddt'
}

TEST_CONTENT = {"a": "b"}


class MockResponse:  # pylint: disable=too-few-public-methods
    def __init__(self, text: str, status_code: int):
        self.text = text
        self.status_code = status_code

    # def text(self):
    #     return self._text


def mocked_requests_get(
        url: str,
        headers: Optional[Dict[str, str]] = None,
) -> MockResponse:
    # pylint: disable=unused-argument
    name, language = url.split('/')[-1].split('+')
    name = name.split("?")[0]
    if name == 'testdicaaddt' and language == 'ca':
        return MockResponse(dumps(TEST_DICT), 200)
    if name == 'testdicaaddt' and language == 'en':
        return MockResponse(dumps(TEST_DICT), 200)
    return MockResponse("Not found", 404)


def mocked_requests_post(*args: Any, **kwargs: Any) -> MockResponse:
    # pylint: disable=unused-argument
    data = loads(kwargs['data'])
    name = data['name']
    language = data['language']

    if name == 'testdicaaddt' and language == 'ca':
        return MockResponse(dumps({"status": "created"}), 200)
    if name == 'testdicaaddt' and language == 'en':
        response = {"message": "Dictionary " + name + " already exists in language " + language}
        return MockResponse(dumps(response), 400)
    return MockResponse(dumps({}), 404)


# Our test case class
class TestDictionary(unittest.TestCase):
    @patch('soil.api.requests.get', side_effect=mocked_requests_get)
    def test_get_dictionary(self, mock_get: MagicMock) -> None:

        response = dictionary('testdicaaddt', 'ca')
        self.assertEqual(response, TEST_DICT)
        response = dictionary('testdicaaddt', 'en')
        self.assertEqual(response, TEST_DICT)

        self.assertEqual(len(mock_get.call_args_list), 2)

    # TO DO: Test create new dict.
    @patch('soil.api.requests.post', side_effect=mocked_requests_post)
    def test_create_dictionary(self, mock_post: MagicMock) -> None:

        response = dictionary('testdicaaddt', 'ca', TEST_CONTENT)
        self.assertEqual(response, {"status": "created"})

        response = dictionary('testdicaaddt', 'en', TEST_CONTENT)
        self.assertEqual(response, {"status": 'Dictionary testdicaaddt already exists in language en'})

        self.assertEqual(len(mock_post.call_args_list), 2)



'''
    DataStructure to serialize PredictionsCombinations
    It stores the data to ElasticSearch.
'''

from datetime import datetime
from soil.data_structures.es_data_structure import ESDataStructure


def _row_to_action(index, _metadata):
    def action_fn(index_row):
        _i, row = index_row
        return {
            '_index': index,
            'generated_at': datetime.now(),
            ** row.to_dict()
        }
    return action_fn


class Predictions(ESDataStructure):
    '''
    DataStructure to serialize PredictionsCombinations
    It stores the data to ElasticSearch.
    '''
    def _row_to_action(self, *args):
        # pylint: disable=no-self-use
        return _row_to_action(*args)



''' Defines soil.data()'''
from typing import Any, Dict, Optional
from soil.api import get_result, upload_data, get_alias
from soil.data_structure import DataStructure, _get_data_structure_name_and_serialize


def data(data_object: Any, metadata: Optional[Dict[str, Any]] = None) -> DataStructure:
    ''' Load data from the cloud or mark it as uploadable '''
    if isinstance(data_object, str):
        # Data object is an id or an alias
        try:
            data_object = _load_data_alias(data_object)
        except KeyError:
            pass
        return _load_data_id(data_object)
    return _upload_data(data_object, metadata)


def _upload_data(data_object: Any, metadata: Optional[Dict[str, Any]] = None) -> DataStructure:
    ds_name, serialized = _get_data_structure_name_and_serialize(data_object)
    result = upload_data(ds_name, serialized, metadata)
    ds = DataStructure(result['_id'], dstype=result['type'])
    return ds


def _load_datastructure(did: str, dtype: str) -> DataStructure:
    # TODO dynamically load a data structure
    return DataStructure(did, dstype=dtype)


def _load_data_alias(alias: str) -> str:
    return get_alias(alias)['state']['result_id']


def _load_data_id(data_id: str) -> DataStructure:
    result = get_result(data_id)
    return _load_datastructure(result['_id'], result['type'])



# pylint: disable=missing-docstring,no-self-use,line-too-long

import unittest
from unittest.mock import patch, call, MagicMock
from typing import NamedTuple, Optional, Dict
from json import dumps
from urllib.parse import urlparse, parse_qs
import numpy  # type: ignore
import pandas  # type: ignore
from soil.data_structure import DS_NAMESPACE
import soil


class MockHttpResponse(NamedTuple):
    """Soil configuration class"""

    status_code: int
    text: str


# pylint: disable=unused-argument
def mock_http_post(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[Dict[str, str]] = None,
) -> MockHttpResponse:
    assert json is not None
    response = {"_id": "mock_id", "type": json["type"]}
    return MockHttpResponse(status_code=201, text=dumps(response))


# pylint: disable=unused-argument, too-many-return-statements
def mock_http_get(
    url: str, headers: Optional[Dict[str, str]] = None
) -> MockHttpResponse:
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    if (
        url_parts.path == "/v2/states/"
        and query_params["name"][0] == "non_existing_alias_but_valid_id"
    ):
        return MockHttpResponse(status_code=200, text=dumps([]))
    if (
        url_parts.path == "/v2/states/"
        and query_params["name"][0] == "invalid_alias_invalidid"
    ):
        return MockHttpResponse(status_code=200, text=dumps([]))
    if url_parts.path == "/v2/states/" and query_params["name"][0] == "valid_alias":
        return MockHttpResponse(status_code=200, text=dumps([{"_id": "alias_id"}]))
    if url_parts.path == "/v2/states/alias_id/":
        result = {"state": {"result_id": "non_existing_alias_but_valid_id"}}
        return MockHttpResponse(status_code=200, text=dumps(result))
    if url_parts.path == "/v2/results/non_existing_alias_but_valid_id/":
        return MockHttpResponse(
            status_code=200,
            text=dumps({"_id": "mock_id", "type": "data_structures.some.test.type"}),
        )
    if url_parts.path == "/v2/results/invalid_alias_invalidid/":
        return MockHttpResponse(
            status_code=404,
            text=dumps(
                {"message": "Document with id invalid_alias_invalidid not found"}
            ),
        )
    raise Exception("mock http case not found")


@patch("soil.api.requests.get", side_effect=mock_http_get)
@patch("soil.api.requests.post", side_effect=mock_http_post)
class TestData(unittest.TestCase):
    def test_upload_data_object_list(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data([1, 2, 3, 4])
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "list.List"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            headers={"Authorization": "Bearer test_token"},
            json={
                "type": "soil.data_structures.predefined.list.List",
                "data": "[1, 2, 3, 4]",
                "metadata": None,
            },
        )  # NOQA
        assert mock_get.call_count == 0

    def test_upload_data_object_dict(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data({"test": "dictionary", "value": 11})
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "dict.Dict"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            headers={"Authorization": "Bearer test_token"},
            json={
                "type": "soil.data_structures.predefined.dict.Dict",
                "data": '{"test": "dictionary", "value": 11}',
                "metadata": None,
            },
        )  # NOQA
        assert mock_get.call_count == 0

    def test_upload_data_object_numpy(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data(numpy.array([[1, 2], [3, 4]]))
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "ndarray.Ndarray"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            headers={"Authorization": "Bearer test_token"},
            json={
                "type": "soil.data_structures.predefined.ndarray.Ndarray",
                "data": "[[1, 2], [3, 4]]",
                "metadata": None,
            },
        )  # NOQA
        assert mock_get.call_count == 0

    def test_upload_data_object_pandas(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data(pandas.DataFrame(data={"col1": [1, 2], "col2": [3, 4]}))
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "pd_data_frame.PdDataFrame"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            headers={"Authorization": "Bearer test_token"},
            json={
                "type": "soil.data_structures.predefined.pd_data_frame.PdDataFrame",
                "data": '{"col1":{"0":1,"1":2},"col2":{"0":3,"1":4}}',
                "metadata": None,
            },
        )  # NOQA
        assert mock_get.call_count == 0

    def test_get_data_valid_id(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data("non_existing_alias_but_valid_id")
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == "data_structures.some.test.type"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        assert mock_post.call_count == 0
        assert mock_get.call_count == 2
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=non_existing_alias_but_valid_id&with-roles",
                    headers={"Authorization": "Bearer test_token"},
                ),
                call(
                    "http://test_host.test/v2/results/non_existing_alias_but_valid_id/",
                    headers={"Authorization": "Bearer test_token"},
                ),
            ]
        )

    def test_get_data_valid_alias(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data("valid_alias")
        data = soil.data("non_existing_alias_but_valid_id")
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == "data_structures.some.test.type"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        assert mock_post.call_count == 0
        assert mock_get.call_count == 5
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=valid_alias&with-roles",
                    headers={"Authorization": "Bearer test_token"},
                ),
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    headers={"Authorization": "Bearer test_token"},
                ),
                call(
                    "http://test_host.test/v2/results/non_existing_alias_but_valid_id/",
                    headers={"Authorization": "Bearer test_token"},
                ),
                call(
                    "http://test_host.test/v2/states/?name=non_existing_alias_but_valid_id&with-roles",
                    headers={"Authorization": "Bearer test_token"},
                ),
                call(
                    "http://test_host.test/v2/results/non_existing_alias_but_valid_id/",
                    headers={"Authorization": "Bearer test_token"},
                ),
            ]
        )

    def test_get_data_invalid_alias(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        self.assertRaises(Exception, soil.data, "invalid_alias_invalidid")
        assert mock_post.call_count == 0
        assert mock_get.call_count == 2
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=invalid_alias_invalidid&with-roles",
                    headers={"Authorization": "Bearer test_token"},
                ),
                call(
                    "http://test_host.test/v2/results/invalid_alias_invalidid/",
                    headers={"Authorization": "Bearer test_token"},
                ),
            ]
        )



# pylint: disable=missing-docstring,no-self-use,line-too-long, unused-argument
# type: ignore

import unittest
from unittest.mock import MagicMock
from soil import decorator

train = MagicMock(name='train')
my_model = MagicMock(name='my_model')


def trainable(module):
    def dec(**model_params):
        def train_fn(**train_params):
            def inner_module(*data):
                return train(*data, module=module, model_params=model_params, **train_params)
            return inner_module
        return train_fn
    return dec


class TestDecorator(unittest.TestCase):

    def setUp(self):
        train.reset_mock()
        my_model.reset_mock()

    def test_no_depth(self):
        with self.assertRaises(ValueError):
            decorator()(trainable)
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls(self):
        my_trainable = decorator(depth=3)(trainable)
        assert my_trainable(__show_calls=True) == (trainable, [])
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_2(self):
        my_trainable = decorator(depth=3)(trainable)
        assert my_trainable(__show_calls=True) == (trainable, [])
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_3(self):
        my_trainable = decorator(depth=3)(trainable)(my_model)
        assert my_trainable(__show_calls=True) == (trainable, [((my_model,), {})])
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_4(self):
        my_trainable = decorator(depth=3)(trainable)(my_model)(test=1)
        assert my_trainable(__show_calls=True) == (trainable, [((my_model,), {}), ((), {'test': 1})])
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_5(self):
        my_trainable = decorator(depth=3)(trainable)(my_model)(test=1)(['bla'])
        assert my_trainable(__show_calls=True) == (trainable, [((my_model,), {}), ((), {'test': 1}), ((['bla'],), {})])
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_6(self):
        my_trainable = decorator(depth=3)(trainable)(my_model)(test=1)(test2='3')('other')
        # pylint:disable=protected-access
        assert my_trainable._extract_mock_name() == 'train()'
        train.assert_called_once_with('other', module=my_model, test2='3', model_params={'test': 1})
        my_model.assert_not_called()



"""This module implements SOIL's Command LIne Interface"""
import argparse
import sys
import os
import json
import shutil
import getpass
import zipfile
from typing import Dict, Tuple, Optional
from json import JSONDecodeError  # Flake8...
import subprocess  # nosec
import virtualenv  # type: ignore
import requests
import yaml

parser = argparse.ArgumentParser(prog="soil")
subparsers = parser.add_subparsers(dest="command")

configure_cmd = subparsers.add_parser("configure", help="Configure SOIL-CLI")
configure_cmd.add_argument(
    "--reset", action="store_true", help="Reset the current configuration"
)

login_cmd = subparsers.add_parser("login", help="Login to configured soil instance")
login_cmd.add_argument("--user", type=str, help="User name or email of the user")
login_cmd.add_argument("--password", type=str, help="Password")

init_cmd = subparsers.add_parser(
    "init", help="Initialize a directory as a soil project"
)
init_cmd.add_argument(
    "dirname",
    nargs="*",
    metavar="DIR",
    default=".",
    help="Name of a directory to initialize as a soil app",
)


install_cmd = subparsers.add_parser(
    "install", help="Installs the provided packages to the soil instance."
)
install_cmd.add_argument(
    "--recursive",
    "-r",
    default=False,
    action="store_true",
    help="Recursive. Use a txt file instead.",
)
install_cmd.add_argument(
    "packages",
    nargs="+",
    metavar="PACKAGE[S]",
    help="Package[s] to install or a txt file if --recursive to install from.",
)

run_cmd = subparsers.add_parser(
    "run", help="Runs the provided module in the current soil project"
)
run_cmd.add_argument(
    "chapter",
    metavar="chapter",
    help="Chapter in soil.yml (setup, data, migration, ...)",
)
run_cmd.add_argument("module", metavar="module", help="Module of chapter in soil.yml")
run_cmd.add_argument(
    "module_args", nargs=argparse.REMAINDER, help="Arguments to be passed to the module"
)

test_cmd = subparsers.add_parser("test", help="Launches tests using SOIL. ")
test_cmd.add_argument(
    "test_args",
    nargs=argparse.REMAINDER,
    help="Arguments to be passed to the test command",
)


def save_environment(config: Dict, env: Dict) -> None:
    """
    Saves the environment keys to a file.
    The file is located in ~/.soil/soil.env and ~/.soil/:app_id/soil.env
    Where app_id is the app_id in the config.
    Notice that ~/.soil/soil.env will store the last login only.
    """

    try:
        os.makedirs(os.path.expanduser("~/.soil/" + config["auth_app_id"]))
    except FileExistsError:
        # directory already exists
        pass

    with open(os.path.expanduser("~/.soil/soil.env"), "w", encoding='utf-8') as env_file:
        env_file.write(json.dumps(env, indent=4, sort_keys=True))

    shutil.copyfile(
        os.path.expanduser("~/.soil/soil.env"),
        os.path.expanduser("~/.soil/" + config["auth_app_id"] + "/soil.env")
    )


def get_soil_root(relpath: str):  # type: ignore
    """Checks if the current dir is under a soil environment and returns its root. Returns None otherwise."""

    path = os.path.abspath(relpath) + "/"

    while path != "/":
        path, _ = os.path.split(path)
        if "soil.yml" in os.listdir(path):
            return path

    return None


def soil_init() -> Tuple[Dict, Optional[Dict]]:
    """Loads configuration and environment."""

    project_root = get_soil_root(".")
    try:
        if project_root is None:
            raise FileNotFoundError("Project root not found.")
        with open(project_root + "/soil.conf", encoding='utf-8') as config_file:
            config = json.loads(config_file.read())
    except FileNotFoundError:
        try:
            with open(os.path.expanduser("~/.soil/soil.conf"), "r", encoding='utf-8') as config_file:
                config = json.loads(config_file.read())
        except (IOError, JSONDecodeError):
            if sys.argv[1] != "configure":
                try:
                    os.rename(
                        os.path.expanduser("~/.soil/soil.conf.bak"),
                        os.path.expanduser("~/.soil/soil.conf"),
                    )
                except IOError:
                    print(
                        "Can not load soil configuration. Plase run soil configure to configure it."
                    )
                    sys.exit()
            else:
                config = None
    try:
        with open(os.path.expanduser("~/.soil/soil.env"), "r", encoding='utf-8') as env_file:
            env = json.loads(env_file.read())
        return config, env
    except (IOError, JSONDecodeError):
        if sys.argv[1] != "configure" and sys.argv[1] != "login":
            try:
                os.rename(
                    os.path.expanduser("~/.soil/soil.conf.bak"),
                    os.path.expanduser("~/.soil/soil.conf"),
                )
            except IOError:
                print(
                    "Can not load soil environment. Plase run soil configure to initialize it."
                )
                sys.exit()
        return config, None


def configure(args: argparse.Namespace, config: Dict) -> None:
    """
    Allows the user to provide the configuration parameters interactively and stores them into a file.
    """

    try:
        os.makedirs(os.path.expanduser("~/.soil/"))
    except FileExistsError:
        # directory already exists
        pass

    try:
        if args.reset:
            try:
                os.rename(
                    os.path.expanduser("~/.soil/soil.conf"),
                    os.path.expanduser("~/.soil/soil.conf.bak"),
                )
                os.rename(
                    os.path.expanduser("~/.soil/soil.env"),
                    os.path.expanduser("~/.soil/soil.env.bak"),
                )
            except FileNotFoundError:
                print(
                    "Seems there is no previous configuration. Performing a clean config..."
                )
                args.reset = False
        with open(os.path.expanduser("~/.soil/soil.conf"), "x", encoding='utf-8') as config_file:
            config = {}
            config["soil_url"] = input(
                "Enter url of your soil instance: [https://soil.amalfianalytics.com]"
            ).replace(
                "\t", ""
            )  # nosec - Input is safe in python3
            if config["soil_url"] == "":
                config["soil_url"] = "https://soil.amalfianalytics.com"
            auth_url_msg = (
                "Enter authentication provider URL: [https://auth.amalfianalytics.com] "
            )
            config["auth_url"] = input(auth_url_msg).replace(
                "\t", ""
            )  # nosec - Input is safe in python3
            if config["auth_url"] == "":
                config["auth_url"] = "https://auth.amalfianalytics.com"
            config["auth_app_id"] = input("Enter your application id: ").replace(
                "\t", ""
            )  # nosec - Input is safe in python3
            config["auth_api_key"] = input("Enter your API key: ").replace(
                "\t", ""
            )  # nosec - Input is safe in python3
            config_file.write(json.dumps(config, indent=4, sort_keys=True))

        # Create the environment file
        with open(os.path.expanduser("~/.soil/soil.env"), "x", encoding='utf-8') as env_file:
            env_file.write("{\n}")

        if args.reset:
            os.remove(os.path.expanduser("~/.soil/soil.conf.bak"))
            os.remove(os.path.expanduser("~/.soil/soil.env.bak"))

    except FileExistsError:
        if not args.reset:
            print("Soil already configured (Use --reset to reconfigure):\n")
            for option, value in config.items():
                print(option + ": " + value)


def login(args: argparse.Namespace, config: Dict, env: Optional[Dict]) -> None:
    """
    Authenticates to the authentication backend and stores the credentials (JWT) in the environment
    """

    print(f"Authenticating to {config['auth_url']} for app {config['auth_app_id']}...")

    if env is None:
        env = {}

    if args.user is not None and args.password is not None:
        username = args.user
        password = args.password
    else:
        if "auth" in env:
            username = input(
                "Username: ["
                + env["auth"]["user"].get(
                    "username", env["auth"]["user"].get("email", "")
                )
                + "]"
            )  # nosec - Input is safe in python3
            if username == "":
                username = env["auth"]["user"].get(
                    "username", env["auth"]["user"].get("email", "")
                )
        else:
            username = input("Username: ")  # nosec - Input is safe in python3
        password = getpass.getpass()

    request_json = {
        "loginId": username,
        "password": password,
        "applicationId": config["auth_app_id"],
    }

    resp = requests.post(
        config["auth_url"] + "/api/login",
        headers={"Authorization": config["auth_api_key"]},
        json=request_json,
    )

    # if resp.status_code == 200 or resp.status_code == 202:
    if resp.status_code in {200, 202}:
        env["auth"] = json.loads(resp.content)
        save_environment(config, env)
        print(
            "Successfully logged in as "
            + env["auth"]["user"].get("username", env["auth"]["user"].get("email", ""))
            + "!"
        )
    elif resp.status_code == 404:
        print("The user was not found or the password was incorrect.")
        sys.exit(1)
    elif resp.status_code == 400:
        print(
            f"Invalid parameters. Perhaps the api key {config['auth_api_key']} or \
                 the app id {config['auth_app_id']} are not correct."
        )
        sys.exit(1)
    else:
        print("Login failed with status code", resp.status_code)
        sys.exit(1)


def init(args: argparse.Namespace) -> None:
    """
    Initializes the current directory or the one provided as argument as a SOIL project by placing
    all the directory structure and creating a python virtual environment.
    """

    dirname = vars(args)["dirname"][0]

    if dirname != ".":
        try:
            os.mkdir(dirname)
        except FileExistsError:
            print("The directory " + dirname + " already exists")
            sys.exit()

    if get_soil_root(dirname):
        if dirname != ".":
            os.rmdir(dirname)
        print("This location is already a soil project")
        sys.exit()
    else:
        with zipfile.ZipFile(
            os.path.dirname(os.path.realpath(__file__)) + "/template.zip", "r"
        ) as template_zip:
            template_zip.extractall(os.path.abspath("./" + dirname))
        # copy soil.conf to local directory
        shutil.copyfile(
            os.path.expanduser("~/.soil/soil.conf"),
            os.path.abspath("./" + dirname + "/soil.conf"),
        )
        # create and activate the virtual environment
        virtualenv.cli_run([os.path.abspath("./" + dirname + "/.venv")])


def install(args: argparse.Namespace) -> None:
    """
    Placeholder install to install remote requirements.
    """
    print("This is a placeholder command to install remote requirements.")
    print("Install arguments:", args)
    # soil_root = get_soil_root('.')
    # if not soil_root:
    #     print("This folder is not initalized as a soil project. Please run soil init to initialize it.")
    #     sys.exit()

    # packages = vars(args)["packages"]

    # for package in packages:
    #     subprocess.check_call(['.venv/bin/python', "-m", "pip", "install", package])    # nosec
    # f = open('requirements.txt', 'w')
    # f.write(os.popen(soil_root + '/.venv/bin/pip freeze').read())    # nosec
    # f.close()


def run(args: argparse.Namespace) -> None:
    """
    Runs the script provided as argument using the python in the virtual environment
    """

    soil_root = get_soil_root(".")
    if not soil_root:
        print(
            "This folder is not initalized as a soil project. Please run soil init to initialize it."
        )
        sys.exit()

    chapter = vars(args)["chapter"]
    module = vars(args)["module"]
    module_args = vars(args)["module_args"]

    os.chdir(soil_root)
    with open("soil.yml", encoding='utf-8') as f:
        conf = yaml.safe_load(f)
    try:
        script = [script for script in conf[chapter] if script.get(module) is not None][
            0
        ][module]
    except (IndexError, KeyError):
        print("Script", module, "not found.")
        sys.exit()
    params = script.get("params", {})
    params = [["--" + k, str(v)] for (k, v) in params.items()]
    params = [item for sublist in params for item in sublist]
    try:
        shell = [".venv/bin/python", "-m", script["path"]] + module_args + params
        print("Running:", " ".join(shell))
        subprocess.run(shell, check=True)  # nosec
    except subprocess.CalledProcessError:
        sys.exit()


def test(args: argparse.Namespace) -> None:
    """
    Launches tests in the test folder.
    """
    soil_root = get_soil_root(".")
    if not soil_root:
        print(
            "This folder is not initalized as a soil project. Please run soil init to initialize it."
        )
        sys.exit()

    test_args = vars(args)["test_args"]

    os.chdir(soil_root)
    try:
        subprocess.run([".venv/bin/python", "-m", "unittest"] + test_args, check=True)  # nosec

    except subprocess.CalledProcessError:
        sys.exit()


def main() -> None:
    """main function"""

    args = parser.parse_args()

    config, env = soil_init()

    try:
        if args.command == "configure":
            configure(args, config)
        elif args.command == "login":
            login(args, config, env)
        elif args.command == "init":
            init(args)
        elif args.command == "install":
            install(args)
        elif args.command == "run":
            run(args)
        elif args.command == "test":
            test(args)
    except KeyboardInterrupt:
        print("\nKeyboardInterrupt - Exit...")
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)  # NOQA # pylint: disable=protected-access


if __name__ == "__main__":
    main()



''' Utils library '''


def read_data(input_file):
    ''' read data from a dataset'''
    # TODO: Get this from the schema!
    mapping = {
        "sepal_length": float,
        "sepal_width": float,
        "petal_length": float,
        "petal_width": float,
        "species": lambda x: x,
        "_id": int
    }
    headers = input_file.readline().strip().split(',')
    lines = input_file.readlines()
    data = []
    for index, line in enumerate(lines):
        element = {'_id': index}
        for k, v in zip(headers, line.strip().split(',')):
            element[k] = mapping[k](v.strip())
        data.append(element)
    input_file.close()
    return data


def get_columns_dictnames(config_cols):
    ''' making a dictionary out of the columns '''
    cols_dictnames = {}
    for group in config_cols:
        names = []
        cols_info = config_cols[group]
        for item in cols_info:
            a = cols_info[item]['name']
            names.append(a)
        cols_dictnames[group] = names
    return cols_dictnames



'''
Soil package.
'''
# pylint:disable=consider-using-from-import
from soil.modulify import modulify
from soil.data import data
from soil.alias import alias
from soil import finder
from soil.logger import logger
import soil.connectors as connectors
from soil.decorator import decorator
from soil.task import task, task_wait
from soil.alerts import alerts
from soil.dictionary import dictionary

finder.install()

__all__ = [
    'modulify',
    'data_structures',
    'modules',
    'data',
    'alias',
    'logger',
    'connectors',
    'decorator',
    'task',
    'task_wait',
    'alerts',
    'dictionary'
]



'''
    DataStructure to serialize a Keras Model
    It puts the network architecture as metadata and pickles the weights.
'''
import pickle
from soil.data_structures.data_structure import DataStructure


class RandomForestModel(DataStructure):
    ''' Data Structure for a Sklearn Model '''
    @staticmethod
    def unserialize(serialized, metadata):
        ''' Function to deserialize '''
        return RandomForestModel(pickle.loads(serialized), metadata)

    def serialize(self):
        ''' Function to serialize '''
        return pickle.dumps(self.data)

    def get_data(self, **_args):
        # pylint: disable=no-self-use
        ''' Placeholder function for the API call '''
        return {}  # self.data



''' Types used by soil library'''

from enum import Enum
from typing import List, Dict
from typing_extensions import TypedDict

Plan = List[Dict[str, str]]


class GetModule(TypedDict):
    ''' Type for GET modules/:moduleId '''
    is_package: bool
    public_api: List[str]
    package_type: str


class ExperimentStatuses(Enum):
    ''' Experiment Statuses '''
    WAITING = 'WAITING'
    EXECUTING = 'EXECUTING'
    DONE = 'DONE'
    ERROR = 'ERROR'


class Experiment(TypedDict):
    ''' Type for GET experiments/:experimentId '''
    _id: str
    app_id: str
    outputs: Dict[str, str]
    experiment_status: str


class Result(TypedDict):
    ''' Type for GET results/:resultId '''
    _id: str
    type: str



# pylint: disable=missing-module-docstring, missing-class-docstring, missing-function-docstring
# pylint: disable=no-self-use, protected-access
import unittest
from unittest.mock import MagicMock
from unittest.mock import patch, call
from types import GeneratorType
import pandas
from iris_dataset.data_structures import es_data_structure


class MockDBObject():
    def __init__(self, dbtype=None):
        self.type = dbtype
        self.partial_query = None

    def update_query(self, update):
        self.partial_query = update

    def __eq__(self, other):
        return self.type == other.type and self.partial_query == other.partial_query


class TestESDataStructure(unittest.TestCase):
    def test_unserialize_data_disk(self):
        self.assertRaises(NotImplementedError, es_data_structure.ESDataStructure.unserialize, 'Test', None)

    def test_unserialize_no_db_object(self):
        self.assertRaises(NotImplementedError, es_data_structure.ESDataStructure.unserialize, None, 'something', None)

    @patch.object(es_data_structure.ESDataStructure, '__init__', return_value=None)
    def test_unserialize_db_object_no_partial_query(self, mock_init):
        es_data_structure.ESDataStructure.unserialize(None, {'test': 1}, MockDBObject('ElasticSearch'))
        mdb = MockDBObject('ElasticSearch')
        mdb.partial_query = {}
        assert mock_init.mock_calls == [call(None, {'test': 1}, mdb)]

    def test_unserialize_other_db_object(self):
        self.assertRaises(
            NotImplementedError, es_data_structure.ESDataStructure.unserialize,
            None, 'something', MockDBObject('Invalid DB'))

    def test_db_unserializer(self):
        obj = es_data_structure.ESDataStructure()
        res = obj.db_unserializer([[1, 2, 3, 4], None])
        assert isinstance(res, GeneratorType)
        assert list(res) == [1, 2, 3, 4]

    # def test_get_data(self):
    #     obj = es_data_structure.ESDataStructure()
    #     assert obj.get_data() == {}

    def test_serialize_no_index(self):
        obj = es_data_structure.ESDataStructure()
        self.assertRaises(AttributeError, obj.serialize)

    @patch('iris_dataset.data_structures.es_data_structure.create_db_object')
    @patch('iris_dataset.data_structures.es_data_structure._row_to_action')
    def test_serialize_with_index(self, mock_row_to_action, mock_create_db_object):
        obj = es_data_structure.ESDataStructure()
        obj.metadata = {'index': 'test_index'}
        df = pandas.DataFrame({'test': [1, 2, 3, 4]}).to_dict('records')
        obj.data = df
        obj.serialize()
        assert mock_create_db_object.mock_calls[0] == call(force_rewrite=False, index='test_index')
        name, args, kwargs = mock_create_db_object.mock_calls[1]
        assert name == '().bulk'
        assert kwargs == {}
        assert isinstance(args[0], map)
        actions = list(args[0])
        assert len(actions) == 4
        assert mock_row_to_action.mock_calls[0] == call('test_index', {'index': 'test_index'})
        assert len(mock_row_to_action.mock_calls) == 5
        for (i, d) in enumerate(df):
            assert (d == mock_row_to_action.mock_calls[i + 1][1][0])


    @patch('iris_dataset.data_structures.es_data_structure.create_db_object')
    @patch('iris_dataset.data_structures.es_data_structure._row_to_action')
    def test_serialize_with_index_rewrite(self, mock_row_to_action, mock_create_db_object):
        obj = es_data_structure.ESDataStructure()
        obj.metadata = {'index': 'test_index', 'rewrite': True}
        df = pandas.DataFrame({'test': [1, 2, 3, 4]}).to_dict('records')
        obj.data = df
        obj.serialize()
        assert mock_create_db_object.mock_calls[0] == call(force_rewrite=True, index='test_index')
        name, args, kwargs = mock_create_db_object.mock_calls[1]
        assert name == '().bulk'
        assert kwargs == {}
        assert isinstance(args[0], map)
        actions = list(args[0])
        assert len(actions) == 4
        assert mock_row_to_action.mock_calls[0] == call('test_index', {'index': 'test_index', 'rewrite': True})
        assert len(mock_row_to_action.mock_calls) == 5
        for (i, d) in enumerate(df):
            assert (d == mock_row_to_action.mock_calls[i + 1][1][0])



''' This package contains patching decorators for testing '''
from unittest.mock import patch
import importlib


def patch_modulify(module):
    ''' Patches the modulify from the module in the original class. '''
    def decorator(original_class):
        ''' Patched the modulify from the module in the original class. '''
        orig_setup = original_class.setUp

        # https://stackoverflow.com/a/37890916
        def new_setup(self, *args, **kwargs):
            ''' SetUp function for the tests '''
            # Do cleanup first so it is ready if an exception is raised
            def kill_patches():  # Create a cleanup callback that undoes our patches
                patch.stopall()  # Stops all patches started with start()
                importlib.reload(module)  # Reload our module which restores the original decorator
            # We want to make sure this is run so we do this in addCleanup instead of tearDown
            self.addCleanup(kill_patches)
            # Now patch the decorator where the decorator is being imported from
            # The lambda makes our decorator into a pass-thru. Also, don't forget to call start()
            patch('soil.modulify', lambda *x, **y: lambda f: f).start()
            # HINT: if you're patching a decor with params use something like:
            # lambda *x, **y: lambda f: f
            importlib.reload(module)  # Reloads the uut.py module which applies our patched decorator
            orig_setup(self, *args, **kwargs)
        original_class.setUp = new_setup
        return original_class
    return decorator



''' A generic datastructure to index stuff to Elastic Search '''
import logging
import statistics
from soil.data_structures.streams.stream import Stream
from soil.connectors.elastic_search import create_db_object

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


def _row_to_action(index, _metadata):
    def function(row):
        return {
            '_index': index,
            ** row
        }
    return function


class ESDataStructure(Stream):
    ''' A generic datastructure to index stuff to Elastic Search '''
    @staticmethod
    def unserialize(data, metadata, db_object=None):
        ''' Extract data from ES'''
        if data is not None:
            raise NotImplementedError('ESDataStructure without ES not implemented')
        if db_object is not None and db_object.type == 'ElasticSearch':
            if db_object.partial_query is None:
                db_object.update_query({})
            return ESDataStructure(None, metadata, db_object)
        raise NotImplementedError('ESDataStructure without ES not implemented')

    def db_unserializer(self, data):
        ''' Unserialize data. Method needed by Stream'''
        # pylint: disable=no-self-use
        return (d for d in data[0])

    def get_data(self, query=None, source=None, action=None):
        ''' Serialize the data for JSON the data '''
        if query is None:
            query = {
                "match_all": {}
            }
        if source is not None:
            question = {'query': query, '_source': source}
        else:
            question = {'query': query}

        response = list(self.db_object.query(question)[0])

        if action == 'count':
            return len(response)
        if action == 'source_avg':
            return statistics.mean(i[source] for i in response)

        return response

    def serialize(self):
        ''' Puts the data in Elastic Search '''
        rewrite = self.metadata.get('rewrite', False)
        index_name = self.metadata['index']
        es_db_object = create_db_object(index=index_name, force_rewrite=rewrite)
        data = self.data
        actions = map(_row_to_action(index_name, self.metadata), data)
        es_db_object.bulk(actions)
        return es_db_object



'''
This module overrides Python's import system
https://docs.python.org/3/library/importlib.html
https://stackoverflow.com/a/43573798/3481480
'''

# pylint: disable=no-self-use,unused-argument,abstract-method,exec-used
import sys
import os.path
from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_file_location, spec_from_loader
from importlib.machinery import ModuleSpec
from typing import List, Optional, Sequence, Union
import types
from soil import api


class Finder(MetaPathFinder):
    '''
    Custom finder that uploads a module or data_structure to the cloud and if
    the module is not found locally it is downloaded from the cloud
    '''
    def find_spec(
            self,
            fullname: str,
            path: Optional[Sequence[Union[bytes, str]]],
            target: Optional[types.ModuleType] = None
    ) -> Optional[ModuleSpec]:
        ''' find_spec implementation '''
        can_be_in_db = False
        if path is None or path == '':
            path = [os.path.abspath(os.path.curdir)]  # top level import --
        if '.' in fullname:
            *parents, name = fullname.split('.')
        else:
            name = fullname
            parents = []
        if len(parents) > 1 and parents[0] == 'soil' and (parents[1] == 'modules' or parents[1] == 'data_structures'):
            path = list(path)
            path.append(os.environ['MODULES_PATH'] if parents[1] == 'modules' else os.environ['DATA_STRUCTURES_PATH'])
            can_be_in_db = True
        else:
            # If the module is not from soil.modules or soil.data_structures use the default loader
            return None
        for entry in path:
            entry = str(entry)
            if os.path.isdir(os.path.join(entry, name)):
                # this module has child modules
                filename = os.path.join(entry, name, '__init__.py')
                submodule_locations: Optional[List[str]] = [os.path.join(entry, name)]
            else:
                filename = os.path.join(entry, name + '.py')
                submodule_locations = None
            if not os.path.exists(filename):
                continue

            return spec_from_file_location(
                fullname,
                filename,
                loader=CustomLoader(fullname, filename),
                submodule_search_locations=submodule_locations
            )

        # If the module is not found look for it in the DB
        if can_be_in_db:
            mod = api.get_module(fullname)
            is_package = mod.get('is_package', False)
            public_api = mod.get('public_api', [])
            package_type = mod.get('package_type', '')
            code = _generate_code(public_api, package_type)
            if code is not None:
                return spec_from_loader(
                    fullname,
                    loader=CustomLoader(fullname, code=code, from_db=True, is_package=is_package),
                    is_package=is_package
                )
        return None  # returning None raises not found error


def _generate_code(public_api: List[str], package_type: str) -> str:
    code = ''
    if package_type == 'modules':
        for f in public_api:
            code += f'from soil import modulify\n@modulify(_from_db=True)\ndef {f}(*args, **kwargs):\n    pass\n\n'
    if package_type == 'data_structures':
        for f in public_api:
            code += f'class {f}:\n    pass\n\n'
    return code


class CustomLoader(Loader):
    ''' This class runs modules and uploads them '''
    def __init__(
            self,
            fullname: str,
            filename: Optional[str] = None,
            code: Optional[str] = None,
            from_db: bool = False,
            is_package: bool = False
    ):
        self.filename = filename
        self.fullname = fullname
        self.code = code
        self.from_db = from_db
        self.is_package = is_package

    def create_module(self, spec: ModuleSpec) -> None:
        ''' Use default module creation semantics '''
        return None

    def exec_module(self, module: types.ModuleType) -> None:
        ''' Load code if necessary, upload it and run it. '''
        is_package = self.is_package
        if self.code is None and self.filename is not None:
            is_package = self.filename[-len('__init__.py'):] == '__init__.py'
            with open(self.filename, encoding='utf-8') as f:
                self.code = f.read()
        if self.code is not None and (
                self.fullname[:len('soil.modules')] == 'soil.modules'
                or self.fullname[:len('soil.data_structures')] == 'soil.data_structures'
        ):
            if not self.from_db and os.environ.get('PY_ENV', 'development') != 'test':
                api.upload_module(self.fullname, self.code, is_package)
            name = self.filename if self.filename is not None else self.fullname
            compiled_module = compile(self.code, name, 'exec')
            exec(compiled_module, vars(module))  # nosec


def install() -> None:
    '''Inserts the finder into the import machinery'''
    sys.meta_path.insert(0, Finder())



''' Package for Elastic Search Connector '''
from typing import Iterable
from typing import Any, Optional


class ElasticSearchDBObject:
    ''' Represents an Elastic Search Data Base Object. '''
    # pylint: disable=unused-argument,redefined-builtin
    def bulk(self, actions: Iterable[Any]) -> Any:
        ''' Performs actions in bulk as defined in the ES bulk API.'''

    def query(self, query: Any, random_sorting: bool = False) -> Any:
        ''' Launches a query to ES. Returns the query.'''

    def insert(self, document: Any, id: Optional[str] = None) -> Any:
        ''' Inserts a document to ES. '''

    def update_query(self, new_partial_query: Any) -> Any:
        ''' Updates a query without executing it.'''


def create_db_object(index: str, force_rewrite: bool = False) -> ElasticSearchDBObject:
    ''' Obtains an Elastic Search DB object. Only works on cloud.'''
    # pylint: disable=unused-argument
    return ElasticSearchDBObject()



''' Module to train a predictor. '''
import argparse
import logging
import soil
from soil.modules.retrain_predictor import retrain_predictor


logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


def train_predictor():
    ''' Main function to train a predictor. '''
    data = soil.data('iris')
    predictor_ref, = retrain_predictor(data)
    soil.alias('predictor', predictor_ref)


def main():
    ''' Argument parsing. '''
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    train_predictor(**vars(args))


if __name__ == '__main__':
    main()



''' Module for BaseStorage '''
from dataclasses import dataclass
from typing import Type, TypeVar, Dict, Any

StorageClass = TypeVar('StorageClass')


@dataclass
class BaseStorage:
    '''
    Abstract class that implements serialize and deserialize methods for storage classes.
    '''
    def serialize(self) -> Dict[str, Any]:
        '''
        Serializes the storage object. In general it shouldn't be used from a module.
        '''

        # TODO if we upgrade to python 3.8 we could use TypedDict
        # https://stackoverflow.com/a/54198204

    @classmethod
    def deserialize(cls: Type[StorageClass], serialized_storage_object: Dict[str, Any]) -> StorageClass:
        '''
        Takes a serialized storage object and returns an instance.
        '''



'''
This package contains calls for the alerts system.
'''
from typing import Any, Dict
from soil import api


def event(key: str, value: Any) -> Any:
    '''
    Creates an event.
    '''
    return api.create_event(key, value)


def alert(alert_config: Dict) -> Any:
    '''
    Creates an alert.
    '''
    return api.create_alert(alert_config)



''' This program runs when new data arrives. '''
import logging
import argparse
import json
import soil
from soil.modules.to_es_data_structure import to_es_data_structure
from ..lib.utils import get_columns_dictnames, read_data

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


def _new_data(input_file=None, schema_file=None):

    # Read schema
    schema = json.load(schema_file)
    columns = get_columns_dictnames(schema['columns'])

    # Read data
    data = read_data(input_file)
    data_ref = soil.data(data, metadata={'columns': columns, 'index': 'simple-data'})
    res, = to_es_data_structure(data_ref)
    soil.alias('iris', res)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=argparse.FileType('r'), help='The input file')
    parser.add_argument('--schema-file', type=argparse.FileType('r'),
                        help='The file with the schema of the data', required=True)
    args = parser.parse_args()
    _new_data(**vars(args))



''' Module to do the predictions given a model and data points to predict on. '''
from soil import modulify
# from soil.data_structures.es_data_structure import Predictions
from soil.data_structures.predictions import Predictions


@modulify(output_types=lambda *input_types, **args: [Predictions])
def make_predictions(model, unlabeled):
    ''' Module function to do the predictions given a model and data points to predict on. '''
    regressor = model.data

    x_labels = model.metadata['x_labels']

    d_y = regressor.predict(unlabeled.data[x_labels])

    new_df = unlabeled.data[x_labels].copy()
    new_df['species'] = d_y
    new_df_dict = new_df.to_dict('records')
    predictions = list({'_id': i, **v} for i, v in enumerate(new_df_dict))

    return [Predictions(predictions, metadata={'index': 'simple-preds'})]



''' This module defines some useful functions '''
import random
from typing import List, Dict, Any, Callable


def _isalambda(value: Any) -> bool:
    return callable(value) and value.__name__ == "<lambda>"


def generate_data_structure_ids(fn_name: str, quantity: int) -> List[str]:
    ''' Generates an id for a datastructure '''
    rnd = random.randint(1000, 9999)  # nosec
    return [fn_name.replace('.', '/') + '-' + str(rnd) + '-' + str(i) for i in range(quantity)]


def build_function(function: Callable) -> Dict[str, Any]:
    '''
    Mounts the argument structure to serialize a function. It uses the reserved dict key `__soil_arg_type`
    to mark a function. The function code must be inside modules or data_structures folders.
    It recognises 3 types of functions:
        * Named functions: `__soil_arg_type='function'` functions with a module and a name.
        * Lambda functions: `__soil_arg_type='lambda'` They are not actually serialized but kept as a placeholder
            for the pattern matching in the server. They can only be used as a decorator parameter.
        * Decorated functions: `__soil_arg_type='decorated'` A function that has been decorated with one or more
            decorators that have been decorated with soil.decorator or functions returned by that decorated function.
    '''

    first_qual = function.__qualname__.split('.')[0]
    if function.__module__ == 'soil.decorator' and first_qual == 'decorator':
        wrapped, calls = function(__show_calls=True)
        return {
            '__soil_arg_type': 'decorated',
            'module': wrapped.__module__,
            'name': wrapped.__name__,
            'calls': [{
                'args': [build_arguments(arg) for arg in call[0]],
                'kwargs': build_arguments(call[1]),
            } for call in calls]
        }

    if _isalambda(function):
        return {'__soil_arg_type': 'lambda'}

    return {
        '__soil_arg_type': 'function',
        'module': function.__module__,
        'name': function.__name__
    }


def build_arguments(args: Any) -> Any:
    ''' Transforms the module arguments into serializable arguments '''
    if isinstance(args, dict):
        return {key: build_arguments(value) for (key, value) in args.items()}

    if isinstance(args, list):
        return [build_arguments(value) for value in args]

    if callable(args):
        return build_function(args)

    return args


def generate_transformation(
        input_ids: List[str],
        output_ids: List[str],
        fn_name: str,
        args: Dict[str, Any]
) -> Dict[str, Any]:
    ''' Mounts the dictionary for a module in a plan '''
    tid = fn_name + '-' + str(random.randint(1000, 9999))  # nosec

    return {
        'id': tid,
        'module': fn_name,
        'inputs': input_ids,
        'outputs': output_ids,
        'args': build_arguments(args)
    }



''' A module to create a generic datastructure to index stuff to Elastic Search '''

from soil import modulify
from soil.data_structures.es_data_structure import ESDataStructure


@modulify(output_types=lambda *input_types, **args: [ESDataStructure])
def to_es_data_structure(input_data):
    ''' A module to create a generic datastructure to index stuff to Elastic Search '''
    return [ESDataStructure(input_data.data, metadata=input_data.metadata)]



''' Module to train a predictor. '''
import argparse
import logging
import soil
import pandas as pd
from soil.modules.make_predictions import make_predictions


logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


def get_preditictions(unlabeled_data_file=None):
    ''' Main function to train a predictor. '''
    model = soil.data('predictor')
    data = pd.read_csv(unlabeled_data_file)
    data_ref = soil.data(data)
    preds, = make_predictions(model, data_ref)
    soil.alias('preds', preds)


def main():
    ''' Argument parsing. '''
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--unlabeled-data-file', type=argparse.FileType('r'),
        help='Pass unlabeled data here', required=True
    )
    args = parser.parse_args()
    get_preditictions(**vars(args))


if __name__ == '__main__':
    main()



# pylint: disable=missing-module-docstring, missing-class-docstring, missing-function-docstring, no-self-use

import unittest
from unittest.mock import patch, call
from iris_dataset.setup import reset_dbs


class MockDataRef:
    def __init__(self, alias, metadata=None):
        # pylint: disable=unused-argument
        self.alias = alias
        if alias == 'employees-combinations':
            self.metadata = {'date_colname': 'Date'}

    def __eq__(self, other):
        return self.alias == other.alias


def soil_data_side_effect(alias, metadata=None):
    return MockDataRef(alias, metadata=metadata)


class TestResetDBs(unittest.TestCase):
    @patch('iris_dataset.setup.reset_dbs.soil.alias')
    @patch('iris_dataset.setup.reset_dbs.to_es_data_structure')
    @patch('iris_dataset.setup.reset_dbs.soil.data')
    def test_reset(
        self, soil_data_mock, to_es_data_structure, _mock_alias
    ):
        # pylint: disable=protected-access
        soil_data_mock.side_effect = soil_data_side_effect
        to_es_data_structure.side_effect = lambda ref: [MockDataRef('predictor')]
        reset_dbs._reset_dbs()

        expected_soil_data_calls = [call([], metadata={'index': 'simple-data', 'rewrite': True}),
                                    call([], metadata={'index': 'simple-preds', 'rewrite': True})]
        soil_data_mock.assert_has_calls(expected_soil_data_calls)



""" Module for the compound storage """
from dataclasses import dataclass
from typing import Dict, List, Tuple
from soil.storage.base_storage import BaseStorage


@dataclass
class CompoundStorage(BaseStorage):
    """
        A meta storage composed of other storages as a dict.

        Example:
            compound = CompoundStorage(storages={
                "db": Elasticsearch(index="index1"),
                "disk": ObjectStorage()
            })
            compound["second_db"] = ElasticSearch(index="index2")
            compound["db"].search(query=myquery)
    """

    storages: Dict[str, BaseStorage]

    def __getitem__(self, storage_name: str) -> BaseStorage:
        """
        Return the storage with that storage_name
        """

    def __setitem__(self, storage_name: str, storage: BaseStorage) -> None:
        """
        Set storage to that storage_name
        """

    def __len__(self) -> int:
        """
        Returns the number of storages
        """

    def items(self) -> List[Tuple[str, BaseStorage]]:
        """
        Returns an iterable of sotrage_name, storage tuples
        """



''' Package for soil.task '''
from typing import Callable, Any, List
from soil.data_structure import DataStructure


def task(modulified_module: Callable) -> Callable:
    '''
    Decorates a modulified module in soil. This function is to call
    soil modules from other soil modules.
    Example:
        task(my_module)(data, arg1='val1')
    '''
    def decorator(*args: DataStructure, **kwargs: Any) -> List[DataStructure]:
        return modulified_module(*args, **kwargs)
    return decorator


def task_wait(futures: Any) -> Any:
    '''
    Wait until computation completes and gather results.

    Accepts a future, nested container of futures, iterator, or queue.
    The return type will match the input type.
    Example:
        result = tasks_wait([future1, future2])
    '''
    return futures



''' This program runs when new data arrives. '''
import logging
import argparse
import soil
from soil.modules.to_es_data_structure import to_es_data_structure

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


def _reset_dbs():
    logging.info('Cleaning index: simple-data')
    data = []
    data_ref = soil.data(data, metadata={'index': 'simple-data', 'rewrite': True})
    res, = to_es_data_structure(data_ref)
    soil.alias('iris', res)
    logging.info('Cleaning index: simple-preds')
    data_ref = soil.data(data, metadata={'index': 'simple-preds', 'rewrite': True})
    res, = to_es_data_structure(data_ref)
    soil.alias('preds', res)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    _reset_dbs(**vars(args))



''' This module defines a Pipeline. '''
# To prevent Pipeline not defined: https://stackoverflow.com/a/49872353/3481480
from __future__ import annotations
from time import sleep
import logging
import datetime
import copy
from typing import Optional, Dict, List, Any
from soil import api
from soil.logger import logger as soil_logger
from soil.types import Plan, ExperimentStatuses, Experiment

# How much should wait between api calls
# Remember ES takes some time to index logs
SLEEP_TIME = 1

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Pipeline:
    ''' A Pipeline stores the transformations and dependencies to obtain certain results. '''

    def __init__(self, plan: Optional[Plan] = None) -> None:
        self.plan = plan if plan is not None else []
        self.experiment: Optional[Experiment] = None
        self.finished = False

    def run(self) -> Dict[str, str]:
        ''' Run the Pipeline (blocking call until the experiment finishes) '''
        if self.finished and self.experiment:
            return self.experiment['outputs']
        if self.experiment is None:
            experiment = api.create_experiment(self.plan)
            self.experiment = experiment
        status = api.get_experiment(experiment['_id'])['experiment_status']
        start_date = datetime.datetime.now().astimezone().isoformat()
        while ExperimentStatuses(status) not in [ExperimentStatuses.DONE, ExperimentStatuses.ERROR]:
            sleep(SLEEP_TIME)
            logs = api.get_experiment_logs(experiment['_id'], start_date)
            _print_logs(logs)
            if len(logs) > 0:
                start_date = logs[0]['date']
            status = api.get_experiment(experiment['_id'])['experiment_status']
        sleep(SLEEP_TIME)
        logs = api.get_experiment_logs(experiment['_id'], start_date)
        _print_logs(logs)
        if ExperimentStatuses(status) == ExperimentStatuses.ERROR:
            raise Exception('Pipeline failed')
        logger.debug('experiment_done: %s', experiment['_id'])
        self.finished = True
        return experiment['outputs']

    def add_transformation(self, transformation: Dict[str, str]) -> Pipeline:
        '''
        Add a new transformation to the Pipeline, returns a new Pipeline
        containing the plan of the old Pipeline plus the transformation.
        '''
        new_plan = self.plan + [transformation]
        return Pipeline(plan=new_plan)

    @staticmethod
    def merge_pipelines(*pipelines: Pipeline) -> Pipeline:
        ''' Merges all the Pipelines passed into a new Pipeline that is returned. '''
        merged_plan: Plan = sum([p.plan for p in pipelines], [])
        return Pipeline(plan=merged_plan)


def _print_logs(logs: List[Dict[str, Any]]) -> None:
    for log in logs[::-1]:
        level = getattr(logging, log['level'], logging.INFO)
        nlog = copy.copy(log)
        del nlog['message']
        soil_logger.log(level, '%s - %s', nlog['date'], log['message'], extra=nlog)



'''
This package contains calls for soil alias data structures
'''
from typing import List, Optional
from soil.api import set_alias
from soil.data_structure import DataStructure


def alias(name: str, data_ref: DataStructure, roles: Optional[List[str]] = None) -> None:
    ''' Set an alias to a data reference'''
    # Force to get an id if there is none yet.
    if data_ref.id is None:
        data_ref.get_id()
    assert isinstance(data_ref.id, str)
    set_alias(name, data_ref.id, roles)



#  pylint:disable=unused-argument
'''
This module defines the @modulify decorator.
'''
from typing import List, Any, Callable, Optional, cast
from soil.data_structure import DataStructure
from soil.pipeline import Pipeline
from soil.decorator import decorator
from soil.utils import generate_data_structure_ids, generate_transformation


@decorator(depth=2)
def modulify(
        _func: Optional[Callable[..., List[DataStructure]]] = None,
        *,
        output_types: Optional[Callable] = None,
        centralize: bool = False,
        distribute: bool = False,
        federating: bool = False,
        num_outputs: int = 1,
        _from_db: bool = False
) -> Callable:
    '''Decorates a function to mark it as a soil module.'''
    def modulify_decorator(fn: Callable[..., List[DataStructure]]) -> Callable:
        fnn = fn if _func is None else _func
        mod_name = fnn.__module__ + '.' + fnn.__name__
        if not _from_db and output_types is None:
            raise ValueError('@modulify: ' + mod_name + ' missing output_types')

        def decorated(*inputs: DataStructure, **kwargs: Any) -> List[DataStructure]:
            if not _from_db and output_types is not None:
                # WARNING this could fail if output_types() checks the type of *inputs
                num_outputs_in = len(output_types(*inputs, **kwargs))
            else:
                num_outputs_in = num_outputs
            new_pipeline = Pipeline.merge_pipelines(*[input.pipeline for input in inputs if input.pipeline is not None])
            output_ids = generate_data_structure_ids(mod_name, num_outputs_in)
            input_ids = [d.id if d.id is not None else d.sym_id for d in inputs]
            for i in input_ids:
                assert isinstance(i, str)
            transformation = generate_transformation(
                input_ids=cast(List[str], input_ids),
                output_ids=output_ids,
                fn_name=mod_name,
                args=kwargs
            )
            new_pipeline = new_pipeline.add_transformation(transformation)
            outputs = [
                DataStructure(sym_id=sym_id, pipeline=new_pipeline)
                for sym_id in output_ids
            ]
            return outputs
        return decorated
    if _func is None:
        return modulify_decorator
    return modulify_decorator(_func)



"""
This package contains calls to the SOIL's REST API
"""
import logging
from typing import Dict, Any, Optional, List
import json
from urllib.parse import urlencode
import requests
from soil.configuration import GLOBAL_CONFIG
from soil.types import Plan, GetModule, Experiment, Result

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

API_ROOT = str(GLOBAL_CONFIG.host) + "/v2/"


def _get_headers() -> dict:
    return {"Authorization": "Bearer " + str(GLOBAL_CONFIG.token)}


def upload_data(dtype: str, data: Any, metadata: Any) -> Result:
    """Upload data to the cloud as a new dataset."""
    logger.debug("upload_data:%s")
    url = API_ROOT + "results/"
    body = {"type": dtype, "data": data, "metadata": metadata}
    response = requests.post(url, headers=_get_headers(), json=body)
    if response.status_code != 201:
        raise ValueError(response.text)
    return json.loads(response.text)


def get_result(result_id: str) -> Dict[str, Any]:
    """Get the result data"""
    logger.debug("get_result:%s", result_id)
    url = API_ROOT + "results/" + result_id + "/"
    response = requests.get(url, headers=_get_headers())
    if response.status_code != 200:
        raise Exception(response.text)
    return json.loads(response.text)


def get_result_data(
    result_id: str, query: Optional[Dict[str, str]] = None
) -> Dict[str, Any]:
    """Get the result data"""
    if query is None:
        query = {}
    logger.debug("get_result:%s %s", result_id, str(query))
    url = API_ROOT + "results/" + result_id + "/data/?" + urlencode(query)
    response = requests.get(url, headers=_get_headers())
    if response.status_code != 200:
        raise Exception(response.text)
    return json.loads(response.text)


def export_result(
    result_id: str, file_path: str, query: Optional[Dict[str, str]] = None
) -> None:
    """Export result and saves it to a file"""
    if query is None:
        query = {}
    logger.debug("export_result:%s %s", result_id, str(query))
    url = API_ROOT + "results/" + result_id + "/export/?" + urlencode(query)
    response = requests.get(url, headers=_get_headers())
    if response.status_code != 200:
        raise Exception(response.text)
    with open(file_path, 'wb') as f:
        f.write(response.content)


def upload_module(module_name: str, code: str, is_package: bool) -> None:
    """Uploads a module"""
    logger.debug("upload_module:%s", module_name)
    url = API_ROOT + "modules/"
    data = {"name": module_name, "code": code, "is_package": is_package}
    response = requests.post(url, headers=_get_headers(), json=data)
    if response.status_code != 200:
        raise Exception(response.text)


def get_module(full_name: str) -> GetModule:
    """Downloads a module"""
    logger.debug("get_module:%s", full_name)
    url = API_ROOT + "modules/" + full_name + "/"
    response = requests.get(url, headers=_get_headers())
    if response.status_code != 200:
        raise Exception(response.text)
    return json.loads(response.text)


def set_alias(alias: str, result_id: str, roles: Optional[List[str]] = None) -> None:
    """Sets an alias for a result. Updates a previous one with the same name."""
    logger.debug("set_alias: %s = %s", alias, result_id)
    obj = {"name": alias, "state": {"alias": alias, "result_id": result_id}}
    if roles is not None:
        obj["roles"] = roles
    try:
        old_alias_id = get_alias(alias)["_id"]
        url = API_ROOT + "states/" + old_alias_id + "/"
        response = requests.patch(url, headers=_get_headers(), json=obj)
    except KeyError:
        url = API_ROOT + "states/"
        response = requests.post(url, headers=_get_headers(), json=obj)
    if response.status_code != 200:
        raise ValueError(f"Failed to create alias {alias}: {response.text}")

    if roles is not None:
        url = API_ROOT + "results/" + result_id + "/"
        response = requests.patch(url, headers=_get_headers(), json={"roles": roles})
        if response.status_code != 200:
            raise ValueError(f"Failed to patch result {result_id}: {response.text}")


def get_alias(alias: str) -> Dict[str, Any]:
    """Gets an alias"""
    logger.debug("get_alias: %s", alias)
    url = API_ROOT + "states/?name=" + alias + "&with-roles"
    response = requests.get(url, headers=_get_headers())
    if response.status_code != 200:
        raise ValueError("Error getting alias:" + response.text)
    aliases = json.loads(response.text)
    if len(aliases) == 0:
        raise KeyError("Alias not found")
    alias_id = aliases[0]["_id"]
    url = API_ROOT + "states/" + alias_id + "/"
    response = requests.get(url, headers=_get_headers())
    if response.status_code != 200:
        raise ValueError("Error getting alias with id " + response.text)
    return json.loads(response.text)


def create_experiment(plan: Plan) -> Experiment:
    """Runs an experiment in SOIL"""
    logger.debug("create_experiment: %s", str(plan))
    url = API_ROOT + "experiments/"
    experiment = {"name": "", "description": "", "plan": plan}
    response = requests.post(
        url, headers=_get_headers(), json={"experiment": experiment}
    )
    if response.status_code != 200:
        raise ValueError("Error creating the experiment:" + response.text)
    return json.loads(response.text)["experiment"]


def get_experiment(experiment_id: str) -> Experiment:
    """Runs an experiment from SOIL"""
    logger.debug("get_experiment: %s", experiment_id)
    url = API_ROOT + "experiments/" + experiment_id + "/"
    response = requests.get(url, headers=_get_headers())
    if response.status_code != 200:
        raise ValueError(
            "Error getting the experiment " + experiment_id + "\n" + response.text
        )
    return json.loads(response.text)


def get_experiment_logs(experiment_id: str, start_date: str) -> Any:
    """Gets logs from a SOIL experiment"""
    logger.debug("get_experiment_id: %s since %s", experiment_id, start_date)
    url = (
        API_ROOT
        + "experiments/"
        + experiment_id
        + "/logs/?"
        + urlencode({"start_date": start_date})
    )
    response = requests.get(url, headers=_get_headers())
    if response.status_code != 200:
        raise ValueError(
            "Error getting the experiment " + experiment_id + "\n" + response.text
        )
    return json.loads(response.text)


def create_event(key: str, value: Any) -> Any:
    """Saves an event in soil"""
    url = API_ROOT + "alerts/events/"
    response = requests.post(
        url, headers=_get_headers(), json={"key": key, "value": value}
    )
    if response.status_code != 201:
        raise ValueError("Error saving the event:" + response.text)
    return json.loads(response.text)


def create_alert(alert: Dict) -> Any:
    """Creates an alert"""
    url = API_ROOT + "alerts/alerts/"
    response = requests.post(url, headers=_get_headers(), json=alert)
    if response.status_code != 201:
        raise ValueError("Error creating the alert:" + response.text)
    return json.loads(response.text)


def get_dictionary(name: str, language: str) -> Dict[str, Any]:
    """Get the a dictionary"""
    url = API_ROOT + "dictionaries/" + name + "?language+" + language
    response = requests.get(url, headers=_get_headers())
    # return response
    if response.status_code != 200:
        raise ValueError("Error getting dictionary:" + response.text)
    return json.loads(response.text)


def create_dictionary(name: str, language: str, content: Dict) -> Dict[str, Any]:
    """Create a dictionary"""
    logger.debug("create_dictionary: %s", str(name))
    url = API_ROOT + "dictionaries/"
    data = json.dumps(
        {
            "name": name,
            "language": language,
            "content": content,
        }
    )

    headers = _get_headers()
    headers["Content-Type"] = "application/json"
    response = requests.post(url, headers=_get_headers(), data=data)
    # return response
    if response.status_code == 400:
        return {"status": json.loads(response.text)["message"]}
    if response.status_code != 200:
        raise ValueError("Error creating the dictionary:" + response.text)
    return json.loads(response.text)



# pylint: disable=missing-docstring,no-self-use,line-too-long, unused-argument, redefined-outer-name
# type: ignore

import unittest
from unittest.mock import patch
from soil import decorator, modulify, task
from soil.data_structure import DataStructure
import soil


@modulify(output_types=lambda *input_types, **args: [input_types[0]])
def train(*data, module=None, model_params=None, **kwargs):
    res, = task(module)(*data, **model_params)
    return [res]


@decorator(depth=3)
def trainable(module):
    def decorator(**model_params):
        def train_fn(**train_params):
            def inner_module(*data):
                return train(*data, module=module, model_params=model_params, **train_params)
            return inner_module
        return train_fn
    return decorator


@trainable
@modulify(output_types=lambda *input_types, **args: [input_types[0]])
def my_model(*data, **my_model_params):
    if 'p3' in my_model_params:
        return task(my_model_params['p3'](cross_validation=True))(*data)
    return data


@modulify(output_types=lambda *input_types, **args: [input_types[0]])
def my_model2(*data, **my_model_params):
    return data


def test_fn1():
    pass


def test_fn2():
    pass


def mock_soil_data_fn(data):
    return DataStructure('test_id')


class TestUtils(unittest.TestCase):
    @patch('soil.data')
    def test_build_arguments_1(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': [1, 2, 3]}])
        res, = my_model(test_param=1)(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected_args = {
            'module': {
                '__soil_arg_type': 'decorated',
                'module': 'soil.modulify',
                'name': 'modulify',
                'calls': [
                    {
                        'args': [],
                        'kwargs': {'output_types': {'__soil_arg_type': 'lambda'}}
                    }, {
                        'args': [{
                            '__soil_arg_type': 'function',
                            'module': 'unit.soil.test_utils', 'name': 'my_model'
                        }],
                        'kwargs': {}}
                ]
            },
            'model_params': {'test_param': 1},
            'cross_validation': True
        }
        assert first['module'] == 'unit.soil.test_utils.train'
        assert args == expected_args

    @patch('soil.data')
    def test_build_arguments_2(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=[test_fn1, test_fn2])(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": [
                    {
                        "__soil_arg_type": "function",
                        "module": "unit.soil.test_utils",
                        "name": "test_fn1"
                    },
                    {
                        "__soil_arg_type": "function",
                        "module": "unit.soil.test_utils",
                        "name": "test_fn2"
                    }
                ]
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected

    @patch('soil.data')
    def test_build_arguments_3(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=my_model2)(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [],
                            "kwargs": {
                                "output_types": {
                                    "__soil_arg_type": "lambda"
                                }
                            }
                        },
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "function",
                                    "module": "unit.soil.test_utils",
                                    "name": "my_model2"
                                }
                            ],
                            "kwargs": {}
                        }
                    ],
                    "module": "soil.modulify",
                    "name": "modulify"
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected

    @patch('soil.data')
    def test_build_arguments_4(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=trainable(my_model2))(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            }
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model2"
                                                }
                                            ],
                                            "kwargs": {}
                                        }
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify"
                                }
                            ],
                            "kwargs": {}
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable"
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected

    @patch('soil.data')
    def test_build_arguments_5(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=my_model)(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            }
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model"
                                                }
                                            ],
                                            "kwargs": {}
                                        }
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify"
                                }
                            ],
                            "kwargs": {}
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable"
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected

    @patch('soil.data')
    def test_build_arguments_6(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=my_model(f1=2, t1=test_fn1))(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            }
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model"
                                                }
                                            ],
                                            "kwargs": {}
                                        }
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify"
                                }
                            ],
                            "kwargs": {}
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "function",
                                    "module": "unit.soil.test_utils",
                                    "name": "test_fn1"
                                }
                            }
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable"
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected

    @patch('soil.data')
    def test_build_arguments_7(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=my_model(f1=2, t1=test_fn1)(another=1))(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            }
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model"
                                                }
                                            ],
                                            "kwargs": {}
                                        }
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify"
                                }
                            ],
                            "kwargs": {}
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "function",
                                    "module": "unit.soil.test_utils",
                                    "name": "test_fn1"
                                }
                            }
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "another": 1
                            }
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable"
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected

    @patch('soil.data')
    def test_build_arguments_8(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=trainable(my_model2)(f1=2, t1=test_fn1)(another=1))(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            }
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model2"
                                                }
                                            ],
                                            "kwargs": {}
                                        }
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify"
                                }
                            ],
                            "kwargs": {}
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "function",
                                    "module": "unit.soil.test_utils",
                                    "name": "test_fn1"
                                }
                            }
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "another": 1
                            }
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable"
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected

    @patch('soil.data')
    def test_build_arguments_9(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=trainable(my_model2)(f1=2, t1=my_model(test=1))
                        (another=1))(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            }
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model2"
                                                }
                                            ],
                                            "kwargs": {}
                                        }
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify"
                                }
                            ],
                            "kwargs": {}
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "decorated",
                                                    "calls": [
                                                        {
                                                            "args": [],
                                                            "kwargs": {
                                                                "output_types": {
                                                                    "__soil_arg_type": "lambda"
                                                                }
                                                            }
                                                        },
                                                        {
                                                            "args": [
                                                                {
                                                                    "__soil_arg_type": "function",
                                                                    "module": "unit.soil.test_utils",
                                                                    "name": "my_model"
                                                                }
                                                            ],
                                                            "kwargs": {}
                                                        }
                                                    ],
                                                    "module": "soil.modulify",
                                                    "name": "modulify"
                                                }
                                            ],
                                            "kwargs": {}
                                        },
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "test": 1
                                            }
                                        }
                                    ],
                                    "module": "unit.soil.test_utils",
                                    "name": "trainable"
                                }
                            }
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "another": 1
                            }
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable"
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected

    @patch('soil.data')
    def test_build_arguments_10(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{'test': 'test'}])
        res, = my_model(test_param=my_model(f1=2, t1=my_model(test=1))
                        (another=1))(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first['args']
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            }
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model"
                                                }
                                            ],
                                            "kwargs": {}
                                        }
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify"
                                }
                            ],
                            "kwargs": {}
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "decorated",
                                                    "calls": [
                                                        {
                                                            "args": [],
                                                            "kwargs": {
                                                                "output_types": {
                                                                    "__soil_arg_type": "lambda"
                                                                }
                                                            }
                                                        },
                                                        {
                                                            "args": [
                                                                {
                                                                    "__soil_arg_type": "function",
                                                                    "module": "unit.soil.test_utils",
                                                                    "name": "my_model"
                                                                }
                                                            ],
                                                            "kwargs": {}
                                                        }
                                                    ],
                                                    "module": "soil.modulify",
                                                    "name": "modulify"
                                                }
                                            ],
                                            "kwargs": {}
                                        },
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "test": 1
                                            }
                                        }
                                    ],
                                    "module": "unit.soil.test_utils",
                                    "name": "trainable"
                                }
                            }
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "another": 1
                            }
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable"
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {
                            "output_types": {
                                "__soil_arg_type": "lambda"
                            }
                        }
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model"
                            }
                        ],
                        "kwargs": {}
                    }
                ],
                "module": "soil.modulify",
                "name": "modulify"
            }
        }
        assert args == expected



# pylint: disable=missing-module-docstring
import os
from os.path import join, basename

package_name = basename(os.path.dirname(os.path.realpath(__file__)))

os.environ['MODULES_PATH'] = join(os.getcwd(), package_name, 'modules')
os.environ['DATA_STRUCTURES_PATH'] = join(os.getcwd(), package_name, 'data_structures')



''' Module to train the predictor '''
from sklearn.ensemble import RandomForestClassifier
import pandas as pd
from soil import modulify
from soil import logger
from soil.data_structures.random_forest_model import RandomForestModel


@modulify(output_types=lambda *input_types, **args: [RandomForestModel])
def retrain_predictor(data):
    ''' Trains the predictor with available data '''
    model = RandomForestClassifier(oob_score=True)
    metadata = data.metadata

    # #  This is here to prevent a weird bug.
    if data.data is not None:
        data = data.data

    x_labels = metadata['columns']['X_columns']
    y_labels = metadata['columns']['y_columns']

    df = pd.DataFrame(data)
    x = df[x_labels]
    y = df[y_labels].values.ravel()

    model = model.fit(x, y)

    logger.info('Model oob score: %s', model.oob_score_)
    logger.info('Model error: %s', 1 - model.oob_score_)

    model_metadata = {
        'oob_score': model.oob_score_,
        'error': 1 - model.oob_score_,
        'x_labels': x_labels,
        'y_labels': y_labels
        }
    return [RandomForestModel(model, model_metadata)]



''' This module implements a Data Structure '''
import os
import json
from typing import Optional, Dict, Any, Union, Tuple
from soil import api
from soil.pipeline import Pipeline
try:
    import numpy  # type: ignore
    NUMPY_ENABLED = True
except ModuleNotFoundError:
    NUMPY_ENABLED = False
try:
    import pandas  # type: ignore
    PANDAS_ENABLED = True
except ModuleNotFoundError:
    PANDAS_ENABLED = False


DS_NAMESPACE = 'soil.data_structures.predefined.'


class DataStructure:
    ''' Data Structure class'''

    def __init__(
            self,
            did: Optional[str] = None,
            sym_id: Optional[str] = None,
            data: Optional[str] = None,
            metadata: Dict[str, Any] = None,
            pipeline: Optional[Pipeline] = None,
            dstype: Optional[str] = None
    ):
        self.id = did  # Result id
        self.sym_id = sym_id  # Symbolic id used in the pipeline
        self.data = data
        self.metadata = metadata
        self.pipeline = pipeline
        self.dstype = dstype

    def get_data(self, **kwargs: Dict[str, Any]) -> Dict[Union[str, int], Any]:
        ''' Invoke the get_data() method from the data structure in the cloud.'''
        result_id = self.get_id()
        result = api.get_result_data(result_id, query={k: json.dumps(v) for k, v in kwargs.items()})
        self.metadata = result['metadata']
        self.dstype = result['type']
        return _unserialize_data(result['type'], result['results'])

    def export(self, path: str, **kwargs: Dict[str, Any]) -> str:
        '''
            Export a result to a file. The file will be stored in the folder
            and returns the file_path.
        '''
        result_id = self.get_id()
        file_path = os.path.join(path, 'result.zip')
        api.export_result(result_id, file_path, **kwargs)
        return file_path

    def get_id(self) -> str:
        ''' Invoke the get_data() method from the data structure in the cloud.'''
        result_id = self.id
        if result_id is None:
            if not self.pipeline:
                raise AssertionError('Data structure has no pipeline to run.')
            results = self.pipeline.run()  # run
            assert isinstance(self.sym_id, str)
            result_id = results[self.sym_id]
            self.id = result_id
        return result_id

    def __getattribute__(self, name: str) -> Any:
        if name in ('data', 'metadata'):
            if (object.__getattribute__(self, 'data') and name == 'data') or \
                    (object.__getattribute__(self, 'metadata') and name == 'metadata'):
                return object.__getattribute__(self, name)
            result_id = self.id
            if not self.pipeline and result_id is None:
                raise AssertionError('Data structure has no pipeline to run.')
            if result_id is None:
                assert isinstance(self.pipeline, Pipeline)
                assert isinstance(self.sym_id, str)
                results = self.pipeline.run()
                result_id = results[self.sym_id]
                self.id = result_id
            if name == 'data':
                result = api.get_result_data(result_id)
                self.dstype = result['type']
                self.data = _unserialize_data(result['type'], result['results'])
            elif name == 'metadata':
                result = api.get_result(result_id)
                self.dstype = result['type']
            self.dstype = result['type']
            self.metadata = result['metadata']
        return object.__getattribute__(self, name)


def _unserialize_data(dtype: str, data: Any) -> Any:
    if dtype == 'data_structures.predefined.ndarray.Ndarray':
        return numpy.array(data)
    if dtype == 'data_structures.predefined.pd_data_frame.PdDataFrame':
        return pandas.DataFrame(data=data)
    return data


def _get_data_structure_name_and_serialize(data_object: Any) -> Tuple[str, str]:
    t = type(data_object)
    if t == dict:
        return DS_NAMESPACE + 'dict.Dict', json.dumps(data_object)
    if t == list:
        return DS_NAMESPACE + 'list.List', json.dumps(data_object)
    if NUMPY_ENABLED and t == numpy.ndarray:
        return DS_NAMESPACE + 'ndarray.Ndarray', json.dumps(data_object.tolist())
    if PANDAS_ENABLED and t == pandas.core.frame.DataFrame:
        return DS_NAMESPACE + 'pd_data_frame.PdDataFrame', data_object.to_json()
    raise ValueError(f'Unrecognised type: {t}')
